from pathlib import Path
from uuid import uuid4

import pytest

try:
    import spacy  # type:ignore

    has_spacy = bool(spacy)
except ModuleNotFoundError:
    has_spacy = False

from .. import (
    BasePath,
    BlobStat,
    BucketClientFS,
    FluidPath,
    Pathy,
    clear_fs_cache,
    use_fs,
    use_fs_cache,
)
from .conftest import ENV_ID, TEST_ADAPTERS


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_is_path_instance(with_adapter: str) -> None:
    blob = Pathy(f"{with_adapter}://fake/blob")
    assert isinstance(blob, Path)


def test_pathy_error_invalid_scheme_paths() -> None:
    # Initializing a path with an absolute system path warns
    with pytest.raises(ValueError):
        Pathy("c:\\temp\\other\\file.txt")
    with pytest.raises(ValueError):
        Pathy("/tmp/other/file.txt")


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_fluid(with_adapter: str, bucket: str) -> None:
    path: FluidPath = Pathy.fluid(f"{with_adapter}://{bucket}/{ENV_ID}/fake-key")
    assert isinstance(path, Pathy)
    path = Pathy.fluid("foo/bar.txt")
    assert isinstance(path, BasePath)
    path = Pathy.fluid("/dev/null")
    assert isinstance(path, BasePath)
    path = Pathy.fluid("C:\\Windows\\Path")
    assert isinstance(path, BasePath)


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_path_to_local(with_adapter: str, bucket: str) -> None:
    root: Pathy = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/to_local")
    foo_blob: Pathy = root / "foo"
    foo_blob.write_text("---")
    assert isinstance(foo_blob, Pathy)
    use_fs_cache()

    # Cache a blob
    cached: Path = Pathy.to_local(foo_blob)
    second_cached: Path = Pathy.to_local(foo_blob)
    assert isinstance(cached, Path)
    assert cached.exists() and cached.is_file(), "local file should exist"
    assert second_cached == cached, "must be the same path"
    assert second_cached.stat() == cached.stat(), "must have the same stat"

    # Cache a folder hierarchy with blobs
    complex_folder = root / "complex"
    for i in range(3):
        folder = f"folder_{i}"
        for j in range(2):
            gcs_blob: Pathy = complex_folder / folder / f"file_{j}.txt"
            gcs_blob.write_text("---")

    cached_folder: Path = Pathy.to_local(complex_folder)
    assert isinstance(cached_folder, Path)
    assert cached_folder.exists() and cached_folder.is_dir()

    # Verify all the files exist in the file-system cache folder
    for i in range(3):
        folder = f"folder_{i}"
        for j in range(2):
            iter_blob: Path = cached_folder / folder / f"file_{j}.txt"
            assert iter_blob.exists()
            assert iter_blob.read_text() == "---"

    clear_fs_cache()
    assert not cached.exists(), "cache clear should delete file"


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_stat(with_adapter: str, bucket: str) -> None:
    path = Pathy("fake-bucket-1234-0987/fake-key")
    with pytest.raises(ValueError):
        path.stat()
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/stat/foo.txt")
    path.write_text("a-a-a-a-a-a-a")
    stat = path.stat()
    assert isinstance(stat, BlobStat)
    assert stat.size is not None and stat.size > 0
    assert stat.last_modified is not None and stat.last_modified > 0
    with pytest.raises(ValueError):
        assert Pathy(f"{with_adapter}://{bucket}").stat()
    with pytest.raises(FileNotFoundError):
        assert Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/stat/nope.txt").stat()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_resolve(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/fake-key")
    assert path.resolve() == path
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/dir/../fake-key")
    assert path.resolve() == Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/fake-key")


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_exists(with_adapter: str, bucket: str) -> None:
    path = Pathy("./fake-key")
    with pytest.raises(ValueError):
        path.exists()

    # invalid bucket name
    invalid_bucket = "unknown-bucket-name-123987519875419"
    assert Pathy(f"{with_adapter}://{invalid_bucket}").exists() is False
    assert Pathy(f"{with_adapter}://{invalid_bucket}/foo.blob").exists() is False
    # valid bucket with invalid object
    assert Pathy(f"{with_adapter}://{bucket}/not_found_lol_nice.txt").exists() is False
    assert Pathy(f"{with_adapter}://{bucket}/@^@^#$%@#$.txt").exists() is False

    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/directory/foo.txt")
    path.write_text("---")
    assert path.exists()
    for parent in path.parents:
        assert parent.exists()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_glob(with_adapter: str, bucket: str) -> None:
    base = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/")
    root = base / "glob/"
    for i in range(3):
        path = root / f"{i}.file"
        path.write_text("---")
    for i in range(2):
        path = root / f"{i}/dir/file.txt"
        path.write_text("---")

    assert list(root.glob("*.test")) == []
    blobs = sorted(list(root.glob("*.file")))
    assert blobs == [
        root / "0.file",
        root / "1.file",
        root / "2.file",
    ]
    assert list((root / "0").glob("*/*.txt")) == [root / "0/dir/file.txt"]
    assert sorted(base.glob("*lob/")) == [root]
    # Recursive matches
    assert sorted(list(root.glob("**/*.txt"))) == [
        root / "0/dir/file.txt",
        root / "1/dir/file.txt",
    ]
    # rglob adds the **/ for you
    assert sorted(list(root.rglob("*.txt"))) == [
        root / "0/dir/file.txt",
        root / "1/dir/file.txt",
    ]

    bad_root = Pathy(f"{with_adapter}://{bucket}-bad-name")
    assert list(bad_root.glob("*.txt")) == []

    bad_name = root / "@#$%@^@%@#@#$.glob"
    assert list(bad_name.glob("*.glob")) == []


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_unlink_path(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/unlink/")
    path = root / "404.txt"
    with pytest.raises(FileNotFoundError):
        path.unlink()
    path = root / "foo.txt"
    path.write_text("---")
    assert path.exists()
    path.unlink()
    assert not path.exists()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_is_dir(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/is_dir/subfolder/another/my.file")
    path.write_text("---")
    assert path.is_dir() is False
    for parent in path.parents:
        assert parent.is_dir() is True


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_is_dir_bucket(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}")
    assert path.is_dir()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_is_dir_bucket_not_found(with_adapter: str) -> None:
    invalid_bucket = f"unknown-bucket-name-{uuid4().hex[:16]}"
    path = Pathy(f"{with_adapter}://{invalid_bucket}")
    assert path.is_dir() is False


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_is_file(with_adapter: str, bucket: str) -> None:
    path = Pathy(
        f"{with_adapter}://{bucket}/{ENV_ID}/is_file/subfolder/another/my.file"
    )
    path.write_text("---")
    # The full file is a file
    assert path.is_file() is True
    # Each parent node in the path is only a directory
    for parent in path.parents:
        assert parent.is_file() is False


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_iterdir(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/iterdir/")
    # (n) files in a folder
    for i in range(2):
        path = root / f"{i}.file"
        path.write_text("---")

    # 2 files in a subfolder
    (root / "sub/file.txt").write_text("---")
    (root / "sub/file2.txt").write_text("---")

    check = sorted(root.iterdir())
    assert check == [
        root / "0.file",
        root / "1.file",
        root / "sub",
    ]


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_iterdir_pipstore(with_adapter: str, bucket: str) -> None:
    path = Pathy(
        f"{with_adapter}://{bucket}/{ENV_ID}/iterdir_pipstore/prodigy/prodigy.whl"
    )
    path.write_bytes(b"---")
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/iterdir_pipstore")
    res = [e.name for e in sorted(path.iterdir())]
    assert res == ["prodigy"]


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_open_errors(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/open_errors/file.txt")
    # Invalid open mode
    with pytest.raises(ValueError):
        path.open(mode="t")

    # Invalid buffering value
    with pytest.raises(ValueError):
        path.open(buffering=0)

    # Binary mode with encoding value
    with pytest.raises(ValueError):
        path.open(mode="rb", encoding="utf8")


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_open_for_read(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/read/file.txt")
    path.write_text("---")
    with path.open() as file_obj:
        assert file_obj.read() == "---"
    assert path.read_text() == "---"


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_open_for_write(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write/file.txt")
    with path.open(mode="w") as file_obj:
        file_obj.write("---")
        file_obj.writelines(["---"])
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write/file.txt")
    with path.open() as file_obj:
        assert file_obj.read() == "------"


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_open_binary_read(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/read_binary/file.txt")
    path.write_bytes(b"---")
    with path.open(mode="rb") as file_obj:
        assert file_obj.readlines() == [b"---"]
    with path.open(mode="rb") as file_obj:
        assert file_obj.readline() == b"---"
        assert file_obj.readline() == b""


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_readwrite_text(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write_text/file.txt")
    path.write_text("---")
    with path.open() as file_obj:
        assert file_obj.read() == "---"
    assert path.read_text() == "---"


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_readwrite_bytes(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write_bytes/file.txt")
    path.write_bytes(b"---")
    assert path.read_bytes() == b"---"


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_readwrite_lines(with_adapter: str, bucket: str) -> None:
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write_text/file.txt")
    with path.open("w") as file_obj:
        file_obj.writelines(["---"])
    with path.open("r") as file_obj:
        assert file_obj.readlines() == ["---"]
    with path.open("rt") as file_obj:
        assert file_obj.readline() == "---"


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_ls_blobs_with_stat(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/ls")
    for i in range(3):
        (root / f"file_{i}").write_text("NICE")
    files = list(root.ls())
    assert len(files) == 3
    valid_names = [f"file_{i}" for i in range(len(files))]
    for i, blob_stat in enumerate(files):
        assert blob_stat.name in valid_names
        assert blob_stat.size == 4
        assert blob_stat.last_modified is not None


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_owner(with_adapter: str, bucket: str) -> None:
    # Raises for invalid file
    with pytest.raises(FileNotFoundError):
        Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write_text/not_a_valid_blob").owner()

    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/write_text/file.txt")
    path.write_text("---")
    # TODO: How to set file owner to non-None in GCS? Then assert here.
    #
    # NOTE: The owner is always set when using the filesystem adapter, so
    #       we can't assert the same behavior here until we fix the above
    #       todo comment.
    path.owner()
    # dumb assert means we didn't raise
    assert True


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_rename_files_in_bucket(with_adapter: str, bucket: str) -> None:
    # Rename a single file
    from_blob = f"{with_adapter}://{bucket}/{ENV_ID}/rename/file.txt"
    to_blob = f"{with_adapter}://{bucket}/{ENV_ID}/rename/other.txt"
    Pathy(from_blob).write_text("---")
    Pathy(from_blob).rename(to_blob)
    assert not Pathy(from_blob).exists()
    assert Pathy(to_blob).is_file()

    with pytest.raises(FileNotFoundError):
        Pathy(f"{with_adapter}://invlid_bkt_nme_18$%^@57582397/foo").rename(to_blob)


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_rename_files_across_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    # Rename a single file across buckets
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/file.txt").write_text("---")
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/file.txt").rename(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/rename/other.txt"
    )
    assert not Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/file.txt").exists()
    assert Pathy(f"{with_adapter}://{other_bucket}/{ENV_ID}/rename/other.txt").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_rename_folders_in_bucket(with_adapter: str, bucket: str) -> None:
    # Rename a folder in the same bucket
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/folder/one.txt").write_text("---")
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/folder/two.txt").write_text("---")
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/folder/")
    new_path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/other/")
    path.rename(new_path)
    assert not path.exists()
    assert new_path.exists()
    assert Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/other/one.txt").is_file()
    assert Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/other/two.txt").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_rename_folders_across_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    # Rename a folder across buckets
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/folder/one.txt").write_text("---")
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/folder/two.txt").write_text("---")
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rename/folder/")
    new_path = Pathy(f"{with_adapter}://{other_bucket}/{ENV_ID}/rename/other/")
    path.rename(new_path)
    assert not path.exists()
    assert new_path.exists()
    assert Pathy(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/rename/other/one.txt"
    ).is_file()
    assert Pathy(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/rename/other/two.txt"
    ).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_replace_files_in_bucket(with_adapter: str, bucket: str) -> None:
    # replace a single file
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/file.txt").write_text("---")
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/file.txt").replace(
        f"{with_adapter}://{bucket}/{ENV_ID}/replace/other.txt"
    )
    assert not Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/file.txt").exists()
    assert Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/other.txt").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_replace_files_across_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    # Rename a single file across buckets
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/file.txt").write_text("---")
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/file.txt").replace(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/replace/other.txt"
    )
    assert not Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/file.txt").exists()
    assert Pathy(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/replace/other.txt"
    ).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_replace_folders_in_bucket(with_adapter: str, bucket: str) -> None:
    # Rename a folder in the same bucket
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/folder/one.txt").write_text(
        "---"
    )
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/folder/two.txt").write_text(
        "---"
    )
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/folder/")
    new_path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/other/")
    path.replace(new_path)
    assert not path.exists()
    assert new_path.exists()
    assert Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/other/one.txt").is_file()
    assert Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/other/two.txt").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_replace_folders_across_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    # Rename a folder across buckets
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/folder/one.txt").write_text(
        "---"
    )
    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/folder/two.txt").write_text(
        "---"
    )
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/replace/folder/")
    new_path = Pathy(f"{with_adapter}://{other_bucket}/{ENV_ID}/replace/other/")
    path.replace(new_path)
    assert not path.exists()
    assert new_path.exists()
    assert Pathy(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/replace/other/one.txt"
    ).is_file()
    assert Pathy(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/replace/other/two.txt"
    ).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_rmdir(with_adapter: str, bucket: str) -> None:
    blob = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rmdir/one.txt")
    blob.write_text("---")

    # Cannot rmdir a blob
    with pytest.raises(NotADirectoryError):
        blob.rmdir()

    Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rmdir/folder/two.txt").write_text("---")
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rmdir/")
    path.rmdir()
    assert not Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rmdir/one.txt").is_file()
    assert not Pathy(
        f"{with_adapter}://{bucket}/{ENV_ID}/rmdir/other/two.txt"
    ).is_file()
    assert not path.exists()

    # Cannot rmdir an invalid folder
    with pytest.raises(FileNotFoundError):
        path.rmdir()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_samefile(with_adapter: str, bucket: str) -> None:
    blob_str = f"{with_adapter}://{bucket}/{ENV_ID}/samefile/one.txt"
    blob_one = Pathy(blob_str)
    blob_two = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/samefile/two.txt")
    blob_one.touch()
    blob_two.touch()
    assert blob_one.samefile(blob_two) is False

    # accepts a Path-like object
    assert blob_one.samefile(blob_one) is True
    # accepts a str
    assert blob_one.samefile(blob_str) is True


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_touch(with_adapter: str, bucket: str) -> None:
    blob = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/touch/one.txt")
    if blob.is_file():
        blob.unlink()
    # The blob doesn't exist
    assert blob.is_file() is False
    # Touch creates an empty text blob
    blob.touch()
    # Now it exists
    assert blob.is_file() is True

    # Can't touch an existing blob if exist_ok=False
    with pytest.raises(FileExistsError):
        blob.touch(exist_ok=False)

    # Can touch an existing blob by default (no-op)
    blob.touch()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_rglob_unlink(with_adapter: str, bucket: str) -> None:
    files = [
        f"{with_adapter}://{bucket}/{ENV_ID}/rglob_and_unlink/{i}.file.txt"
        for i in range(3)
    ]
    for file in files:
        Pathy(file).write_text("---")
    path = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/rglob_and_unlink/")
    for blob in path.rglob("*"):
        blob.unlink()
    # All the files are gone
    for file in files:
        assert Pathy(file).exists() is False
    # The folder is gone
    assert not path.exists()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_mkdir(with_adapter: str) -> None:
    bucket_name = f"pathy-e2e-test-{uuid4().hex}"
    # Create a bucket
    path = Pathy(f"{with_adapter}://{bucket_name}/")
    path.mkdir()
    assert path.exists()
    # Does not assert if it already exists
    path.mkdir(exist_ok=True)
    with pytest.raises(FileExistsError):
        path.mkdir(exist_ok=False)
    assert path.exists()
    path.rmdir()
    assert not path.exists()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_pathy_ignore_extension(with_adapter: str, bucket: str) -> None:
    """The smart_open library does automatic decompression based
    on the filename. We disable that to avoid errors, e.g. if you
    have a .tar.gz file that isn't gzipped."""
    not_targz = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/ignore_ext/one.tar.gz")
    fixture_tar = Path(__file__).parent / "fixtures" / "tar_but_not_gzipped.tar.gz"
    not_targz.write_bytes(fixture_tar.read_bytes())
    again = not_targz.read_bytes()
    assert again is not None


def test_pathy_raises_with_no_known_bucket_clients_for_a_scheme(
    temp_folder: Path,
) -> None:
    path = Pathy("foo://foo")
    with pytest.raises(ValueError):
        path.client(path)
    # Setting a fallback FS adapter fixes the problem
    use_fs(str(temp_folder))
    assert isinstance(path.client(path), BucketClientFS)


@pytest.mark.skipif(not has_spacy, reason="requires spacy and en_core_web_sm model")
def test_pathy_export_spacy_model(temp_folder: Path) -> None:
    """spaCy model loading is one of the things we need to support"""
    import spacy

    use_fs(temp_folder)
    bucket = Pathy("gs://my-bucket/")
    bucket.mkdir(exist_ok=True)
    model = spacy.blank("en")
    output_path = Pathy("gs://my-bucket/models/my_model")
    model.to_disk(output_path)
    sorted_entries = sorted([str(p) for p in output_path.glob("*")])
    expected_entries = [
        "gs://my-bucket/models/my_model/config.cfg",
        "gs://my-bucket/models/my_model/meta.json",
        "gs://my-bucket/models/my_model/tokenizer",
        "gs://my-bucket/models/my_model/vocab",
    ]
    assert sorted_entries == expected_entries



from datetime import datetime
from pathlib import Path
from typing import List, Union

import typer

from . import BasePath, FluidPath, Pathy
from .about import __version__

app: typer.Typer = typer.Typer(help=f"Pathy command line interface. (v{__version__})")


@app.command()
def cp(from_location: str, to_location: str) -> None:
    """
    Copy a blob or folder of blobs from one bucket to another.
    """
    from_path: FluidPath = Pathy.fluid(from_location)
    if not from_path.exists():
        raise ValueError(f"from_path is not an existing Path or Pathy: {from_path}")
    to_path: FluidPath = Pathy.fluid(to_location)
    if from_path.is_dir():
        to_path.mkdir(parents=True, exist_ok=True)
        for blob in from_path.rglob("*"):
            if not blob.is_file():
                continue
            to_blob = to_path / str(blob.relative_to(from_path))
            to_blob.write_bytes(blob.read_bytes())
    elif from_path.is_file():
        # Copy prefix from the source if the to_path has none.
        #
        # e.g. "cp ./file.txt gs://bucket-name/" writes "gs://bucket-name/file.txt"
        sep: str = to_path._flavour.sep  # type:ignore
        if isinstance(to_path, Pathy) and to_location.endswith(sep):
            to_path = to_path / from_path

        to_path.parent.mkdir(parents=True, exist_ok=True)
        to_path.write_bytes(from_path.read_bytes())


@app.command()
def mv(from_location: str, to_location: str) -> None:
    """
    Move a blob or folder of blobs from one path to another.
    """
    from_path: FluidPath = Pathy.fluid(from_location)
    to_path: FluidPath = Pathy.fluid(to_location)

    if from_path.is_file():
        # Copy prefix from the source if the to_path has none.
        #
        # e.g. "cp ./file.txt gs://bucket-name/" writes "gs://bucket-name/file.txt"
        sep: str = to_path._flavour.sep  # type:ignore
        if isinstance(to_path, Pathy) and to_location.endswith(sep):
            to_path = to_path / from_path
        to_path.parent.mkdir(parents=True, exist_ok=True)
        to_path.write_bytes(from_path.read_bytes())
        from_path.unlink()
        return

    if from_path.is_dir():
        to_path.mkdir(parents=True, exist_ok=True)
        to_unlink: List[Union[Pathy, BasePath, Path]] = []
        for blob in from_path.rglob("*"):
            if not blob.is_file():
                continue
            to_blob = to_path / str(blob.relative_to(from_path))
            to_blob.write_bytes(blob.read_bytes())
            to_unlink.append(blob)
        for unlink in to_unlink:
            unlink.unlink()
        if from_path.is_dir():
            from_path.rmdir()


@app.command()
def rm(
    location: str,
    recursive: bool = typer.Option(
        False, "--recursive", "-r", help="Recursively remove files and folders."
    ),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Print removed files and folders."
    ),
) -> None:
    """
    Remove a blob or folder of blobs from a given location.
    """
    path: FluidPath = Pathy.fluid(location)
    if not path.exists():
        typer.echo(f"rm: {path}: No such file or directory")
        raise typer.Exit(1)

    if path.is_dir():
        if not recursive:
            typer.echo(f"rm: {path}: is a directory")
            raise typer.Exit(1)
        selector = path.rglob("*") if recursive else path.glob("*")
        to_unlink = [b for b in selector if b.is_file()]
        for blob in to_unlink:
            if verbose:
                typer.echo(str(blob))
            blob.unlink()
        if path.exists():
            if verbose:
                typer.echo(str(path))
            path.rmdir()
    elif path.is_file():
        if verbose:
            typer.echo(str(path))
        path.unlink()


@app.command()
def ls(
    location: str,
    long: bool = typer.Option(
        False,
        "--long",
        "-l",
        help="Print long style entries with updated time and size shown.",
    ),
) -> None:
    """
    List the blobs that exist at a given location.
    """
    path: FluidPath = Pathy.fluid(location)
    if not path.exists() or path.is_file():
        typer.echo(f"ls: {path}: No such file or directory")
        raise typer.Exit(1)
    now = datetime.now()
    for blob_stat in path.ls():
        print_name = str(path / blob_stat.name)
        if not long:
            typer.echo(print_name)
            continue
        time_str = ""
        if blob_stat.last_modified is not None:
            then = datetime.fromtimestamp(blob_stat.last_modified)
            if now.year != then.year:
                time_str = "%d %b, %Y"
            else:
                time_str = "%d %b, %H:%M"
            time_str = then.strftime(time_str)
        typer.echo("{0:10}{1:15}{2:10}".format(blob_stat.size, time_str, print_name))


if __name__ == "__main__":
    app()

__all__ = ()



import tempfile
from errno import ENOTDIR
from pathlib import Path
from typing import Any

import pytest

from .. import (
    Blob,
    BlobStat,
    Bucket,
    BucketClient,
    BucketClientFS,
    BucketEntry,
    ClientError,
    Pathy,
    PurePathy,
    use_fs,
)
from ..about import __version__


def test_base_package_declares_version() -> None:
    assert __version__ is not None


def test_base_not_supported(monkeypatch: Any) -> None:
    monkeypatch.setattr(Pathy._flavour, "is_supported", False)
    with pytest.raises(NotImplementedError):
        Pathy()


def test_base_cwd() -> None:
    with pytest.raises(NotImplementedError):
        Pathy.cwd()


def test_base_home() -> None:
    with pytest.raises(NotImplementedError):
        Pathy.home()


def test_base_expanduser() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.expanduser()


def test_base_chmod() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.chmod(0o666)


def test_base_lchmod() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.lchmod(0o666)


def test_base_group() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.group()


def test_base_is_mount() -> None:
    assert not Pathy("gs://fake-bucket/fake-key").is_mount()


def test_base_is_symlink() -> None:
    assert not Pathy("gs://fake-bucket/fake-key").is_symlink()


def test_base_is_socket() -> None:
    assert not Pathy("gs://fake-bucket/fake-key").is_socket()


def test_base_is_fifo() -> None:
    assert not Pathy("gs://fake-bucket/fake-key").is_fifo()


def test_base_is_block_device() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.is_block_device()


def test_base_is_char_device() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.is_char_device()


def test_base_lstat() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.lstat()


def test_base_symlink_to() -> None:
    path = Pathy("gs://fake-bucket/fake-key")
    with pytest.raises(NotImplementedError):
        path.symlink_to("file_name")


def test_base_path_check_mode() -> None:
    tmp_dir = tempfile.mkdtemp()
    root = Pathy.fluid(tmp_dir)

    assert root.is_dir() is True

    # Ignores OSErrors about not a directory and returns false
    def not_dir_fn(mode: int) -> bool:
        err = OSError()
        err.errno = ENOTDIR
        raise err

    assert root._check_mode(not_dir_fn) is False

    # Ignores ValueError
    def value_error_fn(mode: int) -> bool:
        raise ValueError("oops")

    assert root._check_mode(value_error_fn) is False

    # Raises from OSError with unknown code
    def other_os_error_fn(mode: int) -> bool:
        err = OSError()
        err.errno = 1337
        raise err

    with pytest.raises(OSError):
        root._check_mode(other_os_error_fn)

    # Raises other unrelated exceptions
    def other_error_fn(mode: int) -> bool:
        raise BaseException()

    with pytest.raises(BaseException):
        root._check_mode(other_error_fn)


def test_base_path_stat_helpers() -> None:
    tmp_dir = tempfile.mkdtemp()
    root = Pathy.fluid(tmp_dir)

    assert root.is_dir() is True

    file = root / "file.txt"
    file.write_text("hello world")

    assert file.is_file() is True
    assert file.is_dir() is False
    assert file.is_mount() is False
    assert file.is_symlink() is False
    assert file.is_block_device() is False
    assert file.is_char_device() is False
    assert file.is_fifo() is False
    assert file.is_socket() is False

    file.unlink()
    root.rmdir()


def test_pathy_mro() -> None:
    assert PurePathy in Pathy.mro()
    assert Path in Pathy.mro()


def test_path_truediv_operator_overload_with_subclass() -> None:
    class MyPathy(Pathy):
        pass

    custom_pathy = MyPathy("gs://foo/bar")
    base_pathy = Pathy("gs://bar/baz")

    out_pathy: Pathy = base_pathy / custom_pathy
    other_pathy: Pathy = custom_pathy / base_pathy
    assert isinstance(out_pathy, Pathy)
    assert isinstance(other_pathy, Pathy)


def test_buckets_rename_replace(temp_folder: Path) -> None:
    use_fs(temp_folder)
    Pathy.from_bucket("foo").mkdir()
    from_path = Pathy("gs://foo/bar")
    to_path = Pathy("gs://foo/baz")
    # Source foo/bar does not exist
    with pytest.raises(FileNotFoundError):
        from_path.rename(to_path)
    with pytest.raises(FileNotFoundError):
        from_path.replace(to_path)


def test_client_create_bucket(temp_folder: Path) -> None:
    bucket_target = temp_folder / "foo"
    assert bucket_target.exists() is False
    cl = BucketClientFS(temp_folder)
    cl.create_bucket(Pathy("gs://foo/"))
    assert bucket_target.exists() is True


def test_client_base_bucket_raises_not_implemented() -> None:
    bucket: Bucket = Bucket()
    blob: Blob = Blob(bucket, "foo", -1, -1, None, None)
    with pytest.raises(NotImplementedError):
        bucket.copy_blob(blob, bucket, "baz")
    with pytest.raises(NotImplementedError):
        bucket.get_blob("baz")
    with pytest.raises(NotImplementedError):
        bucket.delete_blobs([blob])
    with pytest.raises(NotImplementedError):
        bucket.delete_blob(blob)
    with pytest.raises(NotImplementedError):
        bucket.exists()


def test_client_base_blob_raises_not_implemented() -> None:
    blob: Blob = Blob(Bucket(), "foo", -1, -1, None, None)
    with pytest.raises(NotImplementedError):
        blob.delete()
    with pytest.raises(NotImplementedError):
        blob.exists()


def test_client_base_bucket_client_raises_not_implemented() -> None:
    client: BucketClient = BucketClient()
    with pytest.raises(NotImplementedError):
        client.exists(Pathy("gs://foo"))
    with pytest.raises(NotImplementedError):
        client.is_dir(Pathy("gs://foo"))
    with pytest.raises(NotImplementedError):
        client.get_bucket(Pathy("gs://foo"))
    with pytest.raises(NotImplementedError):
        client.list_blobs(Pathy("gs://foo"))
    with pytest.raises(NotImplementedError):
        client.scandir(Pathy("gs://foo"))
    with pytest.raises(NotImplementedError):
        client.create_bucket(Pathy("gs://foo"))
    with pytest.raises(NotImplementedError):
        client.delete_bucket(Pathy("gs://foo"))


def test_bucket_client_rmdir() -> None:
    client: BucketClient = BucketClient()
    client.rmdir(Pathy("gs://foo/bar"))


def test_bucket_client_get_blob() -> None:
    client: BucketClient = BucketClient()
    assert client.get_blob(Pathy("gs://foo")) is None


def test_bucket_client_make_uri() -> None:
    client: BucketClient = BucketClient()
    assert client.make_uri(Pathy("gs://foo/bar")) == "gs://foo/bar"


def test_client_error_repr() -> None:
    error = ClientError("test_message", 1337)
    assert repr(error) == "(1337) test_message"
    assert f"{error}" == "(1337) test_message"


def test_bucket_entry_defaults() -> None:
    entry: BucketEntry = BucketEntry("name")
    assert entry.is_dir() is False
    assert entry.is_file() is True
    entry.inode()
    assert "BucketEntry" in repr(entry)
    assert "last_modified" in repr(entry)
    assert "size" in repr(entry)
    stat = entry.stat()
    assert isinstance(stat, BlobStat)
    assert stat.last_modified is None
    assert stat.size == -1



from typing import Generator

import pytest

from pathy import BucketClientFS, BucketEntry, Pathy, PathyScanDir, use_fs


class MockScanDir(PathyScanDir):
    def scandir(self) -> Generator[BucketEntry, None, None]:
        yield BucketEntry("test-bucket", is_dir=True, raw=None)


class AbstractScanDir(PathyScanDir):
    def scandir(self) -> Generator[BucketEntry, None, None]:
        return super().scandir()


def test_scandir_abstract_methods(bucket: str) -> None:
    use_fs(True)
    client = BucketClientFS()
    root = Pathy(f"gs://{bucket}/")
    with pytest.raises(NotImplementedError):
        AbstractScanDir(client=client, path=root)  # type:ignore


def test_scandir_custom_class(bucket: str) -> None:
    use_fs(True)
    client = BucketClientFS()
    root = Pathy(f"gs://{bucket}/")
    scandir = MockScanDir(client=client, path=root)
    blobs = [b for b in scandir]
    assert len(blobs) == 1


def test_scandir_next(bucket: str) -> None:
    use_fs(True)
    client = BucketClientFS()
    root = Pathy(f"gs://{bucket}/")
    scandir = MockScanDir(client=client, path=root)
    assert next(scandir.__next__()).name == "test-bucket"  # type:ignore



import json
import os
import random
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any, Generator, Optional, Tuple

import pytest

from pathy import Pathy, set_client_params, use_fs, use_fs_cache

from . import azure_testable, gcs_testable, s3_testable

# Which adapters to use
TEST_ADAPTERS = ["fs"]
if gcs_testable:
    TEST_ADAPTERS.append("gcs")
if s3_testable:
    TEST_ADAPTERS.append("s3")
if azure_testable:
    TEST_ADAPTERS.append("azure")

# A unique identifier used to allow each python version and OS to test
# with separate bucket paths. This makes it possible to parallelize the
# tests.
ENV_ID = f"{sys.platform}.{sys.version_info.major}.{sys.version_info.minor}"


@pytest.fixture()
def bucket() -> str:
    return os.environ.get("PATHY_TEST_BUCKET", "pathy-tests-bucket")


@pytest.fixture()
def other_bucket() -> str:
    return os.environ.get("PATHY_TEST_BUCKET_OTHER", "pathy-tests-bucket-other")


@pytest.fixture()
def temp_folder() -> Generator[Path, None, None]:
    tmp_dir = tempfile.mkdtemp()
    yield Path(tmp_dir)
    shutil.rmtree(tmp_dir)


@pytest.fixture()
def with_fs(temp_folder: Path) -> Generator[Path, None, None]:
    yield temp_folder
    # Turn off FS adapter
    use_fs(False)


def gcs_credentials_from_env() -> Optional[Any]:
    """Extract a credentials instance from the GCS_CREDENTIALS env variable.

    You can specify the contents of a credentials JSON file or a file path
    that points to a JSON file.

    Raises AssertionError if the value is present but does not point to a file
    or valid JSON content."""
    if not gcs_testable:
        return None

    creds = os.environ.get("GCS_CREDENTIALS", None)
    if creds is None:
        return None
    from google.oauth2 import service_account

    json_creds = None
    try:
        json_creds = json.loads(creds)
    except json.decoder.JSONDecodeError:
        pass

    if json_creds is not None:
        fd, path = tempfile.mkstemp()
        try:
            with os.fdopen(fd, "w") as tmp:
                tmp.write(json.dumps(json_creds))
            credentials = service_account.Credentials.from_service_account_file(path)
        finally:
            os.remove(path)
    else:
        # If not a JSON string, assume it's a JSON file path
        credentials = service_account.Credentials.from_service_account_file(creds)
    return credentials


def s3_credentials_from_env() -> Optional[Tuple[str, str]]:
    """Extract an access key ID and Secret from the environment."""
    if not s3_testable:
        return None

    access_key_id: Optional[str] = os.environ.get("PATHY_S3_ACCESS_ID", None)
    access_secret: Optional[str] = os.environ.get("PATHY_S3_ACCESS_SECRET", None)
    if access_key_id is None or access_secret is None:
        return None
    return (access_key_id, access_secret)


def azure_credentials_from_env() -> Optional[str]:
    """Extract an access key ID and Secret from the environment."""
    if not azure_testable:
        return None

    connection_string: Optional[str] = os.environ.get(
        "PATHY_AZURE_CONNECTION_STRING", None
    )
    return connection_string


@pytest.fixture()
def with_adapter(
    adapter: str, bucket: str, other_bucket: str
) -> Generator[str, None, None]:
    tmp_dir = None
    scheme = "gs"
    if adapter == "gcs":
        # Use google-cloud-storage
        use_fs(False)
        credentials = gcs_credentials_from_env()
        if credentials is not None:
            set_client_params("gs", credentials=credentials)
    elif adapter == "s3":
        scheme = "s3"
        # Use boto3
        use_fs(False)
        credentials = s3_credentials_from_env()
        if credentials is not None:
            key_id, key_secret = credentials
            set_client_params("s3", key_id=key_id, key_secret=key_secret)
    elif adapter == "azure":
        scheme = "azure"
        # Use azure-storage-blob
        use_fs(False)
        connection_string = azure_credentials_from_env()
        assert connection_string is not None, "expected valid connection_string in env"
        # Sometimes pass BlobServiceClient, and sometimes a connection string
        if bool(random.getrandbits(1)):
            from azure.storage.blob import BlobServiceClient

            service: BlobServiceClient = BlobServiceClient.from_connection_string(
                connection_string
            )
            set_client_params("azure", service=service)
        else:
            set_client_params("azure", connection_string=connection_string)

    elif adapter == "fs":
        # Use local file-system in a temp folder
        tmp_dir = tempfile.mkdtemp()
        use_fs(tmp_dir)
        bucket_one = Pathy.from_bucket(bucket)
        if not bucket_one.exists():
            bucket_one.mkdir()
        bucket_two = Pathy.from_bucket(other_bucket)
        if not bucket_two.exists():
            bucket_two.mkdir()
    else:
        raise ValueError("invalid adapter, nothing is configured")
    # execute the test
    yield scheme

    if adapter == "fs" and tmp_dir is not None:
        # Cleanup fs temp folder
        shutil.rmtree(tmp_dir)
    use_fs(False)
    use_fs_cache(False)
